import logging
import os
import pdb
from typing import Any, Dict, List

import optuna
from loguru import logger
from optuna.trial import TrialState
from prettytable import PrettyTable
from termcolor import colored

from ..utils.misc import (
    StopWatch,
    create_directories_and_log,
    extract_params_from_config,
)


class LoggerProxy(logging.Handler):  # type: ignore
    """
    A logging handler that redirects logger output to the Loguru logger.
    This class overrides the `emit` method of the `logging.Handler` class to redirect the logger
    output to the Loguru logger, which is highly configurable and provides more advanced logging
    features than the standard Python logging library.
    Attributes:
    - None.
    Methods:
    - emit: Overrides the `emit` method of the `logging.Handler` class to redirect logger
        messages to the Loguru logger.
    Remark on the depth argument:
    - By default, the loguru logger uses a depth of 0, which means that it will show
      information about the calling function (i.e., the function that called the logger).
      Specifying a depth=n>0 will show information about the function that called the calling
      function, and so on, up to a depth of n. This can be useful for debugging complex code
      where multiple layers of function calls are involved.
    - CAUTION: Choosing too large of a depth value could lead to an error if the call stack
      isn’t deep enough to support the specified depth.
    """

    def emit(self, record):
        logger.opt(depth=6, exception=record.exc_info).log(
            record.levelname,  # INFO, WARNING, ERROR, etc.
            record.getMessage(),  # The actual message to log
        )


def redirect_logging_to_loguru():
    """
    Redirects logging-logger output to the Loguru-logger using a LoggerProxy instance.
    """

    def unlock_optuna_logger():
        optuna.logging.disable_default_handler()
        optuna.logging.enable_propagation()

    unlock_optuna_logger()
    logging.root.addHandler(LoggerProxy())


def float_formatter(variable):
    def formatted_float(var):
        return f"{var:.5}"

    if isinstance(variable, float):
        return formatted_float(variable)
    elif isinstance(variable, str):
        stripped_variable = (
            variable.replace(".", "")
            .replace("-", "")
            .replace("e", "")
            .replace("E", "")
            .replace("+", "")
        )
        if stripped_variable.isdigit():
            float_variable = float(variable)
            return formatted_float(float_variable)
        else:
            return variable
    else:
        return variable


class AvoidForbiddenParamsSampler(optuna.samplers.BaseSampler):
    """
    Avoid a certain combination of hyperparameters specified with 'self.forbidden_hyperparams',
    such as 'loss weights all-zero.' The default sampling method is 'TPEsampler', but when the
    forbidden combination of hyperparameters is generated, the method is switched to
    'RandomSampler' to encourage it to select a different parameter.

    Parameters:
    -----------
    forbidden_hyperparams : dict
        A dictionary of forbidden hyperparameters with their corresponding values.

    Methods:
    --------
    -sample_relative(study, trial, search_space):
        Returns a relative sampled point using the TPE sampler.

    -sample_independent(study, trial, param_name, param_distribution):
        Samples an independent point using either the TPE or the Random sampler depending on the
        current hyperparameters set and the forbidden hyperparameters set.

    -infer_relative_search_space(study, trial):
        Infers a relative search space using the TPE sampler.

    -should_use_random_sampler(hyperparams):
        Checks if the current hyperparameters set violates the forbidden hyperparameters set,
        and if so, switches to the Random sampler.

    """

    def __init__(self, forbidden_hyperparams: Dict[str, Dict[str, Any]]):
        # Forbidden combinations specified in a config file
        self.forbidden_hyperparams = forbidden_hyperparams

        # Default sampler
        self.tpe_sampler = optuna.samplers.TPESampler()

        # Random sampler
        self.random_sampler = optuna.samplers.RandomSampler()

    def sample_relative(
        self,
        study: optuna.study.Study,
        trial: optuna.trial.FrozenTrial,
        search_space: Dict[str, Any],
    ) -> Dict[str, Any]:
        return self.tpe_sampler.sample_relative(study, trial, search_space)

    def sample_independent(
        self,
        study: optuna.study.Study,
        trial: optuna.trial.FrozenTrial,
        param_name: str,
        param_distribution: optuna.distributions.BaseDistribution,
    ) -> Any:
        current_param_set = trial.params

        # Try TPEsampler first
        param_value = self.tpe_sampler.sample_independent(
            study, trial, param_name, param_distribution
        )

        current_param_set[param_name] = param_value

        # Check if one of the forbidden combination is generated
        should_use_random_sampler = self.should_use_random_sampler(current_param_set)

        if should_use_random_sampler:
            while True:
                # Encourage selecting different value
                param_value = self.random_sampler.sample_independent(
                    study, trial, param_name, param_distribution
                )
                current_param_set[param_name] = param_value

                if not self.should_use_random_sampler(current_param_set):
                    # Log and exit loop when an acceptable value is selected
                    break
            return param_value
        else:
            return param_value

    def infer_relative_search_space(
        self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial
    ) -> Dict[str, Any]:
        return self.tpe_sampler.infer_relative_search_space(study, trial)

    def should_use_random_sampler(self, hyperparams: Dict[str, Any]) -> bool:
        for label, forbid in self.forbidden_hyperparams.items():
            if all(hyperparams.get(key) == value for key, value in forbid.items()):
                logger.info(
                    colored(
                        f"Forbidden combination of parameter set {label} is selected with TPEsampler. "
                        "Switching to RandomSampler to circumvent...",
                        "yellow",
                    )
                )
                return True
        return False


def run_optuna(objective, device, config):
    """
    Load or create optuna study and run optimization steps.
    Also sets pruner for early-stopping.

    Args:
    - objective(function): a function that train and evaluate a model.
    - config(dict): a dictionary containing hyperparameters.
    """

    def setup_pruner(pruner_name, conf):
        """ """
        logger.info(colored(f"Setting up a {pruner_name} pruner...", "yellow"))
        if "median" in pruner_name.lower():
            return optuna.pruners.MedianPruner(
                n_startup_trials=conf.pruner_startup_trials,
                n_warmup_steps=conf.pruner_warmup_steps,
                interval_steps=conf.pruner_interval_steps,
            )
        elif "hyperband" in pruner_name.lower():
            return optuna.pruners.HyperbandPruner()
        elif "percentile" in pruner_name.lower():
            # 60 percentile pruner is more permissive than median
            return optuna.pruners.PercentilePruner(
                60.0,
                n_startup_trials=conf.pruner_startup_trials,
                n_warmup_steps=conf.pruner_warmup_steps,
                interval_steps=conf.pruner_interval_steps,
            )
        elif "threshold" in pruner_name.lower():
            return optuna.pruners.ThresholdPruner(
                n_warmup_steps=conf.pruner_warmup_steps,
                interval_steps=conf.pruner_interval_steps,
            )
        elif "successivehalving" in pruner_name.lower():
            return optuna.pruners.SuccessiveHalvingPruner()
        elif "none" in pruner_name.lower():
            return optuna.pruners.NopPruner()
        else:
            raise ValueError(f"pruner {pruner_name} is not implemented yet!")

    def start_optimization(objective, conf):
        """
        Load or create study, and start optimization.
        Do not use a particular parameter set.
        """
        num_objectives = (
            len(conf.optimization_candidates)
            if "all" in conf.optimization_target.lower()
            else 1
        )
        pruner = (
            setup_pruner(conf.pruner_name, conf)
            if "tuning" in conf.exp_phase and num_objectives == 1
            else setup_pruner("none", conf)
        )

        study_name = "optuna"
        storage_name = "sqlite:///" + conf.root_dblogs + "/" + study_name + ".db"
        create_directories_and_log(conf.root_dblogs)

        study = optuna.create_study(
            study_name=study_name,
            storage=storage_name,
            load_if_exists=True,
            pruner=pruner,
            directions=["minimize"] * num_objectives,
            sampler=AvoidForbiddenParamsSampler(conf.forbidden_param_sets),
        )

        # wrap the objective inside a lambda and call objective inside it
        # cf.) kaggle.com/general/261870
        study.optimize(
            lambda trial: objective(trial, device, config), n_trials=conf.num_trials
        )
        return study

    def reproduce_trial(objective, conf):
        """
        Start training with the parameter set of the best trial, or that of a specified trial.
        """
        num_objectives = (
            len(conf.optimization_candidates)
            if "all" in conf.optimization_target.lower()
            else 1
        )

        assert "tuning" not in conf.exp_phase, colored(
            "Using the best param is prohibited under tuning phase.", "red"
        )

        assert num_objectives == 1, colored(
            """run with best parameter under multi-objective
                optimization is needed to be implemented...!""",
            "red",
        )

        # Set paths
        study_name = "optuna"
        path_db = conf.root_dblogs + "/" + study_name + ".db"
        storage_name = "sqlite:///" + path_db
        if not os.path.exists(path_db):
            raise ValueError(
                colored(
                    f"{path_db} not found. Hyperparameters could not be restored", "red"
                )
            )

        # Load parameters and start optimization
        study = optuna.load_study(study_name=study_name, storage=storage_name)
        for cnt in range(conf.num_trials):
            trial = (
                study.best_trial
                if "best" in str(conf.reproduce_trial).lower()
                else study.trials[conf.reproduce_trial]
            )
            # report to the console
            logger.info(
                colored(f"reproducing the trial,\n", "yellow")
                + f"{trial},from:\n{path_db}"
            )
            pdb.set_trace()
            paramtable = ConsoleTable("Loaded parameters", ["Parameter name", "Value"])
            for key, value in trial.params.items():
                key = key.replace("space_", "") if key.startswith("space_") else key
                paramtable.add_entry([key, float_formatter(value)])
            paramtable.print()
            # run!
            objective(trial, device, config)
        return study

    def trial_statistics(study, conf):
        """
        Summarize the optimization results.
        """

        def print_value_and_params(trial, conf, num_objectives):
            if num_objectives == 1:
                paramtable = ConsoleTable("Objective", [conf.optimization_target])
                paramtable.add_entry([trial.value])
            else:
                paramtable = ConsoleTable("Objectives", conf.optimization_candidates)
                paramtable.add_entry(trial.values)
            paramtable.print()
            if "tuning" in conf.exp_phase:
                paramtable = ConsoleTable("Best params", ["Parameter name", "Value"])
                for key, value in trial.params.items():
                    key = key.replace("space_", "") if key.startswith("space_") else key
                    paramtable.add_entry([key, float_formatter(value)])
                paramtable.print()

        def print_importance(importances: dict, conf):
            """ """
            if not "tuning" in conf.exp_phase:
                return
            response = input(
                "Would you like to print a parameter importance table? (yes/[no]) "
            )
            if "yes" in response.lower() or "y" in response.lower():
                paramtable = ConsoleTable(
                    "Importance", ["Parameter name", "Percentage"]
                )
                for key, value in importances.items():
                    key = key.replace("space_", "") if key.startswith("space_") else key
                    paramtable.add_entry([key, float_formatter(value * 100)])
                paramtable.print()

        num_objectives = (
            len(conf.optimization_candidates)
            if "all" in conf.optimization_target.lower()
            else 1
        )

        pruned_trials = len(
            study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        )
        completed_trials = len(
            study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])
        )

        valid_trials = pruned_trials + completed_trials
        if valid_trials > 0:
            percent_pruned_trials = 100 * pruned_trials / valid_trials
            logger.info(
                f"{pruned_trials} out of {valid_trials} trials are pruned so far\
                    ({percent_pruned_trials:.2f}%, {conf.pruner_name} pruner). "
            )
        else:
            logger.error("No valid trials found!")

        logger.info("Best trial(s):")
        if num_objectives == 1:
            logger.info(f"Optimization target: {conf.optimization_target}")
            trial = study.best_trial
            print_value_and_params(trial, conf, num_objectives)
            if completed_trials > 1:
                print_importance(optuna.importance.get_param_importances(study), conf)

        else:  # multi-objective
            trials = study.best_trials
            for i, trial in enumerate(trials):
                logger.info("Trial #", i)
                print_value_and_params(trial, conf, num_objectives)
            for i in range(num_objectives):
                if completed_trials > 1:
                    logger.info(
                        f"Optimization target #{i}: {conf.optimization_candidates[i]}"
                    )
                    print_importance(
                        optuna.importance.get_param_importances(
                            study, target=lambda t: t.values[i]
                        ),
                        conf,
                    )

    ########### main ###########
    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "ROOT_DBLOGS",
            "EXP_PHASE",
            "FORBIDDEN_PARAM_SETS",
            "NUM_TRIALS",
            "PRUNER_NAME",
            "PRUNER_STARTUP_TRIALS",
            "PRUNER_WARMUP_STEPS",
            "PRUNER_INTERVAL_STEPS",
            "OPTIMIZATION_CANDIDATES",
            "OPTIMIZATION_TARGET",
            "REPRODUCE_TRIAL",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    redirect_logging_to_loguru()  # from logging logger to loguru logger

    # supported exp_phase:  try, tuning, stat
    assert (
        (conf.exp_phase == "tuning")
        or (conf.exp_phase == "stat")
        or (conf.exp_phase == "try")
    )

    with StopWatch():
        # start optimization
        study = (
            reproduce_trial(objective, conf)
            if conf.reproduce_trial
            else start_optimization(objective, conf)
        )

    # summarize the results at the end
    trial_statistics(study, conf)
    logger.success(colored("Done and dusted!", "yellow"))  # well done!
    return


def format_model_name(model_backbone: str) -> str:
    """ """
    if any(
        char in model_backbone.lower()
        for char in ["lstm", "long short-term memory", "b2bsqrt_tandem"]
    ):
        return "LSTM"
    elif any(
        char in model_backbone.lower()
        for char in ["transformer", "tfmr", "tandemformer"]
    ):
        return "TRANSFORMER"
    elif any(
        char in model_backbone.lower() for char in ["s4", "sfour", "s_four", "ssss"]
    ):
        return "S4"
    else:
        raise ValueError(f"{model_backbone} is not supported!")


class ConsoleTable:
    def __init__(self, title, fields):
        """
        Initializes the ConsoleTable instance with
        the given title and fields.

        Args:
        - title (str): The title of the table.
        - fields (list[str]): The field names of the table.
        """
        self.table = PrettyTable(fields)
        self.table.title = title

    def add_entry(self, entry):
        """
        Adds an entry to the table, which can either be
        a single row or a list of rows.

        Args:
        - entry: The entry to be added to the table.
        """
        if type(entry[0]) == list:
            # add multiple rows to the table.
            for row in entry:
                self.table.add_row(row)
        else:  # add a single row to the table.
            self.table.add_row(entry)

    def print(self):
        """
        Prints the formatted table to the console.
        Also sort by the first entry (e.g., 'Parameter name').
        """
        self.table.sortby = self.table.field_names[0]
        logger.info("\n" + str(self.table))


def sample_parameter(trial, paramname: str, paramspace: dict) -> Any:
    """
    Suggest one parameter from the provided serch space.
    Key values starting "SPACE_" containing parameter search space information.

    Args:
    -paramname (str): a name of parameter search space.
    -paramspace (dict): a python dictionary contains search space info.
        PARAM_SPACE: "float", "int", or "categorical".
        - if float: use suggest_float to suggest a float of range [LOW, HIGH], separated by STEP.
        if LOG=True, a float is sampled from logspace but you shall set STEP=None.
        - if int: use suggest_int to suggest an int of range [LOW, HIGH], separated by STEP.
        STEP should be divisor of the range, otherwise HIGH will be automatically modified.
        if LOG=True, an int is sampled from logspace but you shall set STEP=None.
        - if categorical: use suggest_categorical to select one category from CATEGORY_SET.
        Note that if the parameter is continuous (e.g., 1, 2, 3, ..., or 1.0, 0.1, 0.001, ...),
        it is adviseable to use float or int space because suggest_categorical treats
        each category independently.

    Returns:
    - suggested_parameter: A suggested parameter, either category, float, or int.
    """
    assert type(paramspace) == dict, "SPACE_ must be a dictionary!"

    if "int" in paramspace["PARAM_SPACE"]:
        suggested_parameter = trial.suggest_int(
            paramname,
            low=paramspace["LOW"],
            high=paramspace["HIGH"],
            step=paramspace["STEP"],
            log=paramspace["LOG"],
        )
    elif "float" in paramspace["PARAM_SPACE"]:
        suggested_parameter = trial.suggest_float(
            paramname,
            low=paramspace["LOW"],
            high=paramspace["HIGH"],
            step=paramspace["STEP"],
            log=paramspace["LOG"],
        )
    elif "cat" in paramspace["PARAM_SPACE"]:
        suggested_parameter = trial.suggest_categorical(
            paramname, paramspace["CATEGORY_SET"]
        )
    else:
        raise ValueError(
            'Unknown parameter space! Currently "float",'
            '"int", or "categorical" are supprted.'
        )
    return suggested_parameter


def check_forbidden_parameters(config: dict) -> None:
    """
    Investigate whether the configuration file contains a forbidden combination of
    parameters defined in `config['FORBIDDEN_PARAM_SETS']`.
    """
    for label, paramset in config["FORBIDDEN_PARAM_SETS"].items():
        # Note that the forbidden parameter names are prefixed with 'SPACE_' to indicate the Optuna search space.
        # Thus, the parameter name without the prefix is used as a config key (`param_name[len('SPACE_'):]`).
        if all(
            config[param_name[len("SPACE_") :]] == param_value
            for param_name, param_value in paramset.items()
        ):
            raise ValueError(
                f'Forbidden parameter combination "{label}" is found!'
                f"Please select values other than {paramset.values()}."
            )


def set_config_param(
    key: str, config: Dict[str, Any], trial: optuna.Trial, paramtable: ConsoleTable
) -> str:
    """
    Set the value of a parameter in the config dictionary. See config_definition.py for detail.

    Args:
        key (str): The name of the parameter to be set.
        config (Dict[str, float]): The configuration dictionary where the parameter value will be set.
        trial (Trial): The Optuna trial object that will be used to suggest a new parameter value.
        paramtable (ConsoleTable): An instance of the ConsoleTable class to log the new parameter value.

    Returns:
        str: The new key for the parameter value in the config dictionary.
    """
    param_space = config[key]

    newkey = key.upper().replace("SPACE_", "")

    if config.get(newkey) is not None:
        # suggest and assign the value in-place
        config[newkey] = sample_parameter(trial, key, param_space)
    else:
        raise ValueError(f"{newkey} not found in config!")

    # add to the table
    paramtable.add_entry([newkey, float_formatter(config[newkey])])

    return newkey


def setup_model_parameter_space(
    config: Dict[str, Any], model_key: str, model_zoo: List[str]
) -> None:
    """
    Setup the parameter space for a specific model in the config dictionary.

    Args:
        config (Dict[str, Any]): The config dictionary to be modified.
        model_key (str): The key in the config dictionary representing the name of the model.
        model_zoo (List[str]): A list of available model names.

    Returns:
        None
    """
    # modify the rest of the parameter space
    model_name = config[model_key]
    config.update({key: value for key, value in config[model_name.upper()].items()})
    for key in model_zoo:
        if key.upper() in config:
            del config[key.upper()]


def conditional_suggestion(
    config: Dict[str, Any], trial: optuna.Trial, paramtable: ConsoleTable
) -> None:
    """
    Conditionally suggest hyperparameters using Optuna based on the configuration dictionary.

    Args:
        config (Dict[str, float]): The configuration dictionary containing the hyperparameters to be optimized.
        trial (Trial): The Optuna trial object to be used for hyperparameter optimization.
        paramtable (ConsoleTable): An instance of the ConsoleTable class to log the new parameter values.

    Returns:
        None
    """
    space_is_pe = "SPACE_IS_POSITIONAL_ENCODING"
    space_is_grad = "SPACE_IS_TRAINABLE_ENCODING"
    space_model = "SPACE_MODEL_BACKBONE"
    model_zoo = config[space_model]["CATEGORY_SET"]

    # Sample positional encoding flag
    if space_is_pe in config:
        newkey = set_config_param(space_is_pe, config, trial, paramtable)
        del config[space_is_pe]

        # if positional encoding is used, choose trainable or not
        if config[newkey] and space_is_grad in config:
            _ = set_config_param(space_is_grad, config, trial, paramtable)
        # delete trainable flag space for both pe=True and pe=False case
        del config[space_is_grad]

    # Sample model backbone
    if space_model in config:
        newkey = set_config_param(space_model, config, trial, paramtable)
        del config[space_model]

    # modify config structure according to the suggested model.
    setup_model_parameter_space(config, newkey, model_zoo)


def suggest_parameters(trial: optuna.Trial, config: Dict[str, Any]) -> None:
    """
    Suggest hyperparameters. Note that the keys for hyperparameter lists in config must start with a prefix "SPACE_,"

    Args:
    - trial: A trial object for optuna optimization.
    - config: A dictionary of hyperparameters.

    Returns:
    - config: An updated dictionary with suggested hyperparameters.

    Raise:
    - ValueError if keys in kwargs and config do not match.
    """

    # check if necessary parameters are defined in the config file
    requirements = set(["EXP_PHASE", "FORBIDDEN_PARAM_SETS"])
    conf = extract_params_from_config(requirements, config)

    # run sanity check and return if not sampling is necessary
    if not "tuning" in conf.exp_phase:
        check_forbidden_parameters(config)
        return

    logger.info(colored("Starting a tuning phase.", "yellow"))

    # move parameters that are relevant to current model une level up in the hierarchy

    # create a table for readable output
    paramtable = ConsoleTable(
        "Hyperparameteres suggested by Optuna", ["Parameter name", "Suggested value"]
    )

    # conditional optimization to avoid unnecessary hyperparameters
    conditional_suggestion(config, trial, paramtable)

    # suggest the rest of the hyperparameters
    for key, value in config.items():
        if key.upper().startswith("SPACE_"):
            _ = set_config_param(key, config, trial, paramtable)

    # print a table of suggested params
    paramtable.print()

    return


def report_to_pruner(trial, best, iter, config):
    """
    Report the intermediate "best" variable to decide
    whether to prune the training (early-stopping).

    Args:
    - trial (optuna trial)
    - best (tuple): performance metrics to be minimized.
    - iter (int): the current training step.
    - config (dict): a custom dictionary storeing parameters.

    Returns:
    - None
    """
    # check if necessary parameters are defined in the config file
    requirements = set(["OPTIMIZATION_TARGET"])
    conf = extract_params_from_config(requirements, config)

    if "all" in conf.optimization_target.lower():
        return  # multi-objective pruner is not supported
    else:
        # report intermediate objective value.
        trial.report(*best, iter)
        # handle pruning based on the intermediate value.
        if trial.should_prune():
            raise optuna.TrialPruned()



import torch
import torch.nn.functional as F
from ..utils.data_processing import sequential_slice_labels
from ..utils.performance_metrics import (
    calc_llrs,
    calc_oblivious_llrs,
    seqconfmx_to_macro_ave_sns,
)
from ..utils.misc import extract_params_from_config


def multiplet_crossentropy(logits_slice, labels_slice):
    """Multiplet loss for density estimation of time-series data.
    Args:
        model: A model.backbones_lstm.LSTMModel object.
        logits_slice: An logit Tensor with shape
            ((effective) batch size, order of SPRT + 1, num classes).
            This is the output from LSTMModel.call(inputs, training).
        labels_slice: A label Tensor with shape ((effective) batch size,)
    Returns:
        xent: A scalar Tensor. Sum of multiplet losses.
    """
    # Calc multiplet losses
    logits = logits_slice.permute(1, 0, 2)
    logits = logits.reshape(-1, logits.shape[2])
    labels = labels_slice.repeat(logits_slice.shape[1])
    xent = F.cross_entropy(logits, labels)

    # A scalar averaged in a batch and a sliding window

    return xent


def compute_llrs(logits_concat: torch.Tensor, oblivious: bool = False) -> torch.Tensor:
    """
    Args:
        logits_concat: A logit Tensor with shape
            (batch, (time_steps - order_sprt), order_sprt + 1, num classes).
            This is the output from
            utils.data_processing.sequential_concat(logit_slice, labels_slice)
        oblivious: A bool, whether to use the oblivion formula or not (= TANDEM formula).

    Returns:
        llrs: A Tensor with shape (batch size, time_steps, num cls, num cls).
    """

    if oblivious:
        # Miyagawa and Ebihara, ICML 2021
        llrs = calc_oblivious_llrs(logits_concat)
    else:
        # compute with TANDEM formula: Ebihara+, ICLR 2021
        llrs = calc_llrs(logits_concat)

    return llrs  # (batch_size, time_steps, num cls, num cls)


def compute_loss_for_llr(
    llrs: torch.Tensor, labels_concat: torch.Tensor, version: str
) -> torch.Tensor:
    """
    Compute the loss for log-likelihood ratio estimation (LLLR).
    Args:
        llrs: A Tensor with shape (batch size, time_steps, num cls, num cls).
        labels_concat: A Tensor with shape (batch size,).
        version: A string, which version of the loss to compute.

    Returns:
        loss: A scalar Tensor.
    """
    shapes = llrs.shape
    assert shapes[-1] == shapes[-2]
    time_steps = shapes[1]
    num_classes = shapes[-1]

    labels_oh = F.one_hot(labels_concat, num_classes).to(torch.float32)
    labels_oh = labels_oh.reshape(-1, 1, num_classes, 1)
    # (batch_size, 1, num cls, 1)

    if "LLLR" in version.upper():  # LLLR, Ebihara+, ICLR 2021
        lllr = torch.abs(labels_oh - torch.sigmoid(llrs))
        # (batch, time_steps, num cls, num cls)
        lllr = 0.5 * (num_classes / (num_classes - 1.0)) * torch.mean(lllr)
    elif "LSEL" in version.upper():  # LSEL, Miyagawa and Ebihara, ICML 2021
        llrs = llrs * labels_oh
        # (batch, time_steps, num cls, num cls)
        llrs = torch.sum(llrs, dim=2)
        # (batch, time_steps, num cls)
        llrs = llrs.reshape(-1, num_classes)
        # (batch * time_steps, num cls)
        minllr = torch.min(llrs, dim=1, keepdim=True)[0].detach()
        llrs = llrs - minllr
        # (batch * time_steps, num cls)
        lllr = torch.sum(torch.exp(-llrs), dim=1)
        # (batch * time_steps, )
        lllr = torch.mean(-minllr + torch.log(lllr + 1e-12))
        # scalar
    else:
        raise ValueError(f"version={version} must be either of 'LLLR' or 'LSEL.'")

    return lllr


def extract_shapes(llrs, scores_full):
    llrs_shape = llrs.shape
    batch_size = llrs_shape[0]
    time_steps = llrs_shape[1]
    num_classes = llrs_shape[2]
    num_thresh = scores_full.shape[0]

    return batch_size, time_steps, num_classes, num_thresh


def restore_lost_significance(llrs):
    # Restore lost significance with epsilon
    tri = torch.ones_like(llrs)
    triu = torch.triu(tri)  # Upper triangular part.
    tril = torch.tril(tri)  # Lower triangular part.
    llrs_restore = llrs - 1e-10 * (triu - tril)
    # (batch, time_steps, num cls, num cls)
    # To avoid double hit due to the values exactly equal to 0
    # in scores or when doing truncation, LLRs of the last frame.

    return llrs_restore


def calc_scores(llrs: torch.Tensor, thresh: torch.Tensor) -> torch.Tensor:
    """
    Calculate scores (= LLR - thresh).

    Args:
        llrs: (batch_size, time_steps, num cls, num cls)
        thresh: (num_thresh, batch_size, time_steps, num cls, num cls)
    Returns:
        scores_full: (num_thresh, batch_size, time_steps, num cls, num cls)
    """
    assert len(thresh.shape) == 5, "thresh must be 5-dim tensor."

    llr_mtx = torch.unsqueeze(llrs, dim=0)
    scores_full = llr_mtx - thresh

    return scores_full


def calc_hittimes(llrs, scores_full, num_thresh, time_steps, batch_size, num_classes):
    """ """

    def _calc_truncated_predictions(scores_full, llrs, num_thresh):
        """ """
        scores = torch.min(scores_full, dim=-1)[
            0
        ]  # (num thresh, batch, time_steps, num cls)
        # Calc all predictions and waits
        preds_all = torch.sign(scores) + 1
        # 0:wait, 1:hit (one-hot vector)
        # (num thresh, batch, time_steps, num cls)
        preds_last = torch.sign(torch.min(llrs[:, -1:, :, :], dim=-1)[0]) + 1
        # 0: wait, 1: hit (one-hot vector)
        # (batch, 1, num cls)
        preds_last = torch.unsqueeze(preds_last, dim=0)
        # (1, batch, 1, num cls)
        preds_last = preds_last.repeat(num_thresh, 1, 1, 1)
        # (num thresh, batch, 1, num cls)
        preds_all_trunc = torch.cat([preds_all[:, :, :-1, :], preds_last], dim=2)

        # (num thresh, batch, time_steps - 1, num cls)
        # + (num thresh, batch, 1, num cls)
        # = (num thresh, batch, time_steps, num cls)
        # Now, preds_all_trunc[i, j, t, :] for fixed i and j is
        # a one-hot vector for t = time_steps - 1
        # and
        # filled with 0 or a one-hot vector for t != time_steps - 1.

        return preds_all_trunc

    def _calc_mask(time_steps, num_thresh, batch_size, num_classes, device):
        # For hittimes
        mask = torch.tensor(
            [i + 1 for i in range(time_steps)][::-1], dtype=torch.float32, device=device
        )
        mask = mask.repeat(num_thresh * batch_size * num_classes)
        mask = mask.reshape(num_thresh, batch_size, num_classes, time_steps)
        mask = mask.permute(0, 1, 3, 2)
        # (num thresh, batch, time_steps , num cls)

        return mask

    def _calc_hittimes(preds_all_trunc, mask):
        masked = preds_all_trunc * mask
        # (num thresh, batch, time_steps, num cls)
        hitidx = torch.max(masked, dim=2)[0]
        # (num thresh, batch, num cls)
        hittimes = time_steps - torch.max(hitidx, dim=2)[0] + 1
        # (num thresh, batch)

        assert (
            torch.max(hittimes) <= time_steps
        ), f"hittimes must be \in {1, time_steps}, while got {torch.max(hittimes)=}!"
        return hitidx, hittimes

    _device = llrs.device

    # Calc scores and scores_full
    # scores, scores_full = _calc_scores(llrs, thresh)
    # Score means LLR - thresh.
    # scores:      (num thresh, batch, time_steps, num cls)
    # scores_full: (num thresh, batch, time_steps, num cls, num cls)

    # Calc all predictions with truncation
    preds_all_trunc = _calc_truncated_predictions(scores_full, llrs, num_thresh)
    # (num thresh, batch, time_steps, num cls)
    # Now, preds_all_trunc[i, j, t, :] for fixed i and j is
    # a one-hot vector for t = time_steps - 1
    # and
    # filled with 0 or a one-hot vector for t != time_steps - 1.

    # Calc mean hitting times
    mask = _calc_mask(time_steps, num_thresh, batch_size, num_classes, device=_device)
    # (num thresh, batch, time_steps , num cls)
    hitidx, hittimes = _calc_hittimes(preds_all_trunc, mask)
    # (num thresh, batch)

    return hitidx, hittimes


def calc_ausat(llrs, scores_full, labels_concat):
    """ """

    batch_size, time_steps, num_classes, num_thresh = extract_shapes(llrs, scores_full)

    hitidx, hittimes = calc_hittimes(
        llrs, scores_full, num_thresh, time_steps, batch_size, num_classes
    )
    mht = torch.mean(hittimes, dim=1)
    sns_from_confmx = calc_sat_from_confmx(
        labels_concat, hitidx, num_classes, num_thresh
    )

    return mht, sns_from_confmx


def calc_sat_from_confmx(labels_concat, hitidx, num_classes, num_thresh):
    """
    Calc non-differentiable confusion matrix.
    For visualization purpose.

    Args:
    - labels_concat:
    - hitidx: index of hitting time. Size: (NUM_THRESH, BATCH_SIZE, NUM_CLASSES)

    Return:
    -
    """
    preds = torch.argmax(hitidx, dim=2)
    # (num thresh, batch,)
    preds = F.one_hot(preds, num_classes=num_classes)
    # (num thresh, batch, num cls)

    labels_oh = F.one_hot(labels_concat, num_classes=num_classes)
    # (batch, num cls)
    labels_oh = torch.unsqueeze(labels_oh, dim=0)
    labels_oh = labels_oh.repeat(num_thresh, 1, 1)
    # (num thresh, batch, num cls)
    preds = torch.unsqueeze(preds, dim=-2)
    labels_oh = torch.unsqueeze(labels_oh, dim=-1)
    confmx = torch.sum(labels_oh * preds, dim=1, dtype=torch.int32)
    ls_sns = seqconfmx_to_macro_ave_sns(confmx)

    return ls_sns


def calc_weight_decay(model):
    """ """
    wd_reg = 0.0
    for name, param in model.named_parameters():
        if "bias" not in name:  # Exclude bias terms from weight decay
            wd_reg += torch.norm(param, p=2) ** 2
    return wd_reg


def adaptive_weight_loss(model, losses, constant_weights):
    """ """
    assert (
        len(losses) == model.adaptive_loss_weights.shape[0]
    ), "Number of losses and weights must match!"
    assert (
        len(constant_weights) == model.adaptive_loss_weights.shape[0]
    ), "Number of constant weights and adaptive weights must match!"
    total_loss = 0.0
    for i, (loss, const) in enumerate(zip(losses, constant_weights)):
        weight = torch.finfo(torch.float32).eps + model.adaptive_loss_weights[i] ** 2
        total_loss += 0.5 * loss * const / weight + torch.log(weight)
    return total_loss


def compute_loss_and_metrics(model, x, labels, global_step, config):
    """Calculate loss and gradients.
    Args:
        model: A tf.keras.Model object.
        x: A Tensor. A batch of time-series input data
            without sequential_slice and sequential_concat.
        y: A Tensor. A batch of labels
            without sequential_slice and sequential_concat.
        training: A boolean. Training flag.
        order_sprt: An int. The order of the SPRT.
        time_steps: An int. Num of frames in a sequence.
        num_thresh: An int. Num of thresholds for AUC-SAT.
        beta: A positive float for beta-sigmoid in AUC-SAT.
        param_multiplet_loss: A float. Loss weight.
        param_llr_loss: A float. Loss weight.
        param_aucsat_loss: A float. Loss weight.
        param_wd: A float. Loss weight.
        flag_wd: A boolean. Weight decay or not.
    Returns:
        gradients: A Tensor or None.
        losses: A list of loss Tensors; namely,
            total_loss: A scalar Tensor or 0 if not calc_grad.
                The weighted total loss.
            multiplet_loss: A scalar Tensor.
            llr_loss: A scalar Tensor.
            aucsat_loss: A scalar Tensor.
            wd_reg: A scalar Tensor.
        logits_concat: A logit Tensor with shape
            (batch, (time_steps - order_sprt), order_sprt + 1, num classes).
            This is the output from
            utils.data_processing.sequential_concat(logit_slice, labels_slice).
    Remarks:
        - All the losses below will be calculated if not calc_grad
          for TensorBoard logs.
            total_loss
            multiplet_loss
            llr_loss
            aucsat_loss
            wd_reg
    """
    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "ORDER_SPRT",
            "TIME_STEPS",
            "OBLIVIOUS",
            "NUM_THRESH",
            "SPARSITY",
            "BETA",
            "IS_ADAPTIVE_LOSS",
            "MAX_NORM",
            "BATCH_SIZE",
            "ORDER_SPRT",
            "PARAM_MULTIPLET_LOSS",
            "PARAM_LLR_LOSS",
            "LLLR_VERSION",
            "WEIGHT_DECAY",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    y_slice = sequential_slice_labels(labels, conf.time_steps, conf.order_sprt)

    model_forward = model
    model_module = model

    llrs, logits_slice, thresholds, scores_full = model_forward(x)

    lllr = compute_loss_for_llr(llrs, labels, conf.lllr_version)

    # multiplet_loss
    mce = multiplet_crossentropy(logits_slice, y_slice)

    # AUC
    mht, sns_from_confmx = calc_ausat(llrs, scores_full, labels)

    # L2 weight decay regularization
    wd_reg = calc_weight_decay(model_module)

    if conf.is_adaptive_loss:
        # adaptively change weights with trainable parameters
        total_loss = adaptive_weight_loss(
            model_module,
            [mce, lllr, wd_reg],
            [
                conf.param_multiplet_loss,
                conf.param_llr_loss,
                conf.weight_decay,
            ],
        )

    else:
        # use constant weights predefined in config file
        total_loss = (
            conf.param_multiplet_loss * mce
            + conf.param_llr_loss * lllr
            + conf.weight_decay * wd_reg
        )

    # store loss values
    losses = {
        "total_loss": total_loss,
        "multiplet_crossentropy (MCE)": mce,
        "LLR_estimation (LLRE)": lllr,
        "weight_decay": wd_reg,
    }

    # store performance metrics and training status with the losses
    monitored_values = {
        "losses": losses,
        "thresholds": thresholds,
        "llrs": llrs,
        "mht": mht,
        "sns_from_confmx": sns_from_confmx,
    }

    return monitored_values



import pickle
from types import SimpleNamespace
from typing import Iterator, List, Tuple, Union, cast

import lmdb
import numpy as np
import torch
import torch.distributed as dist
from loguru import logger
from scipy.stats import multivariate_normal
from torch import Tensor
from tqdm import tqdm

from ..utils.misc import extract_params_from_config


def numpy_to_torch(*args):
    """
    Convert numpy arrays to PyTorch tensors.

    Args:
        *args: One or more inputs to be converted.

    Returns:
        A tuple containing the converted inputs.
    """
    converted_args = []
    for arg in args:
        if isinstance(arg, np.ndarray):
            arg = torch.from_numpy(arg)
        if len(args) == 1:
            return arg
        converted_args.append(arg)
    return tuple(converted_args)


def move_to_device(device, *args):
    """
    Convert numpy arrays to PyTorch tensors.

    Args:
        device:
        *args: One or more inputs to be moved to the device.

    Returns:
        A tuple containing the inputs moved to the device..
    """
    moved_args = []
    for arg in args:
        arg = arg.to(device)
        moved_args.append(arg)
    return tuple(moved_args)


def move_data_to_device(
    data: Union[
        Tuple[Tensor, Tensor],
        Tuple[Tensor, Tensor, Tensor],
    ],
) -> Union[Tuple[Tensor, Tensor], Tuple[Tensor, Tensor, Tensor]]:
    """
    Move the input data to the specified device based on the length of the input data tuple.

    Args:
        data: Tuple containing either (x_batch, y_batch, gt_llrs_batch) or (x_batch, y_batch).
        device: The device to move the data to.

    Returns:
        A tuple with data moved to the specified device.

    Raises:
        ValueError: If the length of the input data tuple is not 2 or 3.
    """
    rank = dist.get_rank()

    if len(data) == 3:
        # if the dataset contains ground-truth log likelihood ratio.
        # Casting is required to explicitly show the type to mypy. Note that it does not do any actual runtime checks.
        x_batch, y_batch, gt_llrs_batch = cast(Tuple[Tensor, Tensor, Tensor], data)
        x_batch, y_batch, gt_llrs_batch = move_to_device(
            rank, x_batch, y_batch, gt_llrs_batch
        )
        return x_batch, y_batch, gt_llrs_batch
    elif len(data) == 2:
        # Data and label only: typical real-world data.
        # Casting is required to explicitly show the type to mypy. Note that it does not do any actual runtime checks.
        x_batch, y_batch = cast(Tuple[Tensor, Tensor], data)
        x_batch, y_batch = move_to_device(rank, x_batch, y_batch)
        return x_batch, y_batch
    else:
        raise ValueError(
            "data tuple length is expected either to be "
            f"3 (x, y, llr) or 2 (x, y) but got {len(data)=}!"
        )


class LMDBDataset(torch.utils.data.Dataset):
    """
    A PyTorch dataset for reading data and labels from an LMDB database.

    Attributes:
    - lmdb_path (str): The path to the LMDB database.
    - env (lmdb.Environment): The LMDB environment object.
    - data_size (int): The number of data points in the dataset.

    Methods:
    - __len__: Returns the number of data points in the dataset.
    - __getitem__: Returns the data and label for a given index.

    """

    def __init__(self, lmdb_path: str, names: Tuple[str], is_load_onto_memory=True):
        """
        Required function for PyTorch Dataset class.
        Initializes a new LMDBDataset object.

        Args:
        -lmdb_path (str): The path to the LMDB database.
        -names (Tuple[str]): list of to-be-retrieved data. e.g., ('data', 'label')

        """
        super().__init__()
        self.lmdb_path = lmdb_path
        self.names = names
        self.is_load_onto_memory = is_load_onto_memory

        # Open a read transaction
        with lmdb.open(lmdb_path, readonly=True) as env:
            with env.begin() as txn:
                # Get the total number of data
                self.data_size = txn.stat()["entries"] // len(names)
                if is_load_onto_memory:
                    self.data = []
                    for i in range(self.data_size):
                        item = {}
                        for name in self.names:
                            item_bytes = txn.get(f"{i:08}_{name}".encode("ascii"))
                            item[name] = pickle.loads(item_bytes)
                        self.data.append(item)

    def __len__(self) -> int:
        """
        Required function for PyTorch Dataset class.
        Returns the number of data points in the dataset.
        """
        return self.data_size

    def __getitem__(self, index: int) -> tuple:
        """
        Required function for PyTorch Dataset class.
        Returns the data and label for a given index.

        Args:
            index: The index of the data point.

        Returns:
            A tuple containing the data and label.
        """

        if self.is_load_onto_memory:
            item = self.data[index]
        else:
            # Open the LMDB database for each worker
            with lmdb.open(self.lmdb_path, readonly=True) as env:
                with env.begin(buffers=True) as txn:
                    item = {}
                    for name in self.names:
                        item_bytes = txn.get(f"{index:08}_{name}".encode("ascii"))
                        item[name] = pickle.loads(item_bytes)
        # Convert the data and label to PyTorch tensors
        _tensors = []
        for name in self.names:
            if "label" in name:
                _tensors.append(torch.tensor(item[name]).to(torch.int64))
            else:
                _tensors.append(torch.tensor(item[name]).to(torch.float32))

        return tuple(_tensors)


class LMDBIterableDataset(torch.utils.data.IterableDataset):
    """
    A custom IterableDataset for reading data from an LMDB database.

    Args:
        lmdb_path (str): Path to the LMDB database.
        names (Tuple[str]): A tuple of names representing different data entries.
    """

    def __init__(self, lmdb_path: str, names: Tuple[str]):
        super().__init__()
        self.lmdb_path = lmdb_path
        self.names = names

    def __iter__(self) -> Iterator[Tuple[Tensor, ...]]:
        """
        Iterator that yields data samples from the LMDB database.

        Returns:
            Iterator[Tuple[Tensor, ...]]: An iterator of tuples containing PyTorch tensors.
        """

        # Open the LMDB database for each worker
        env = lmdb.open(self.lmdb_path, readonly=True)

        # Define an iterator that yields data samples
        index = 0
        while True:
            with env.begin(buffers=True) as txn:
                item = {}
                for name in self.names:
                    item_bytes = txn.get(f"{index:08}_{name}".encode("ascii"))

                    # If the item_bytes is None, break the loop
                    if item_bytes is None:
                        break

                    item[name] = pickle.loads(item_bytes)
                else:
                    # Convert the data and label to PyTorch tensors
                    _tensors = []
                    for name in self.names:
                        if "label" in name:
                            _tensors.append(torch.tensor(item[name]).to(torch.int64))
                        else:
                            _tensors.append(torch.tensor(item[name]).to(torch.float32))

                    # Yield the data before incrementing the index
                    yield tuple(_tensors)

                    # Increment index and continue the loop
                    index += 1
                    continue

                # If the loop is broken (no more data), break the outer loop
                break

        # Close the LMDB database for each worker
        env.close()


def write_lmdb(
    lmdb_path: str,
    data: Tuple[np.ndarray],
    names: Tuple[str],
    map_size: int = int(1e12),
) -> None:
    """
    Writes the data and labels to an LMDB database.

    Args:
    - lmdb_path (str): The path to the LMDB database.
    - data (tuple of numpy.ndarray): A tuple of numpy arrays containing the data to be saved.
    - name (tuple of str): A tuple containing the names of the arrays to be saved.
    - map_size (int, optional): The maximum size of the LMDB database in bytes.
                                 Default is 1 terabyte.
    Returns:
    - None
    """
    assert len(data) == len(names), "Number of data and name list must match."

    # Open a new LMDB database
    env = lmdb.open(lmdb_path, map_size=map_size)

    # Get the number of data points
    data_number = data[0].shape[0]
    for data_array in data:
        assert (
            data_array.shape[0] == data_number
        ), f"Total {data_array.shape[0]=} and {data_number=} does not match!"

    # Open a write transaction
    with env.begin(write=True) as txn:
        for i in tqdm(range(data_number)):
            # Write each data array to the database
            for j, data_array in enumerate(data):
                data_bytes = pickle.dumps(data_array[i])
                name = names[j]
                txn.put("{:08}_{}".format(i, name).encode("ascii"), data_bytes)

    # Close the database
    env.close()


def lmdb_dataloaders(config: dict, load_test=False) -> dict:
    """ """
    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "TRAIN_DATA",
            "VAL_DATA",
            "TEST_DATA",
            "DATA_NAMES",
            "BATCH_SIZE",
            "IS_SHUFFLE",
            "NUM_WORKERS",
            "IS_LOAD_ONTO_MEMORY",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    logger.info("loading data... ")
    if conf.is_load_onto_memory:
        logger.info(
            f"If this process takes long, "
            "consider setting is_load_onto_memory=False or use LMDBIterableDataset."
        )

    train_dataset = LMDBDataset(
        conf.train_data,
        names=conf.data_names,
        is_load_onto_memory=conf.is_load_onto_memory,
    )
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=conf.batch_size,
        pin_memory=True,
        num_workers=conf.num_workers,
        shuffle=conf.is_shuffle,
    )

    val_dataset = LMDBDataset(
        conf.val_data,
        names=conf.data_names,
        is_load_onto_memory=conf.is_load_onto_memory,
    )
    val_dataloader = torch.utils.data.DataLoader(
        val_dataset,
        batch_size=conf.batch_size,
        pin_memory=True,
        num_workers=conf.num_workers,
        shuffle=False,
    )

    if load_test:
        test_dataset = LMDBDataset(
            conf.test_data,
            names=conf.data_names,
            is_load_onto_memory=conf.is_load_onto_memory,
        )
        test_dataloader = torch.utils.data.DataLoader(
            test_dataset,
            batch_size=conf.batch_size,
            pin_memory=True,
            num_workers=conf.num_workers,
            shuffle=False,
        )

        data_loaders = {
            "train": train_dataloader,
            "val": val_dataloader,
            "test": test_dataloader,
        }
    else:
        data_loaders = {"train": train_dataloader, "val": val_dataloader}

    return data_loaders


def initialize_multivariate_gaussian(
    conf: SimpleNamespace,
) -> Tuple[np.ndarray, np.ndarray, List]:
    """
    Initialize a multivariate Gaussian distribution for each class.

    Args:
    - conf (SimpleNamespace): an instance of the SimpleNamespace class containing the following keys:
        - num_classes (int): the number of classes.
        - feat_dim (int): the feature dimension.
        - density_offset (float): the density offset used to initialize the mean vectors.

    Returns:
    - meanvecs (ndarray): an array of shape (num_classes, feat_dim) containing the mean vectors for each class.
    - covmat (ndarray): the covariance matrix, which is a diagonal matrix of shape (feat_dim, feat_dim).
    - pdfs (list): a list of multivariate normal distributions, one for each class.
    """

    meanvecs = np.zeros((conf.num_classes, conf.feat_dim))
    covmat = np.eye(conf.feat_dim)
    pdfs = []
    for cls_i in range(conf.num_classes):
        meanvecs[cls_i, cls_i] = conf.density_offset
        pdfs.append(multivariate_normal(meanvecs[cls_i], covmat))
    return meanvecs, covmat, pdfs


def compute_log_likelihood_ratio_matrix(
    x: np.ndarray, pdfs: List, conf: SimpleNamespace
) -> np.ndarray:
    """
    Compute the log-likelihood ratio matrix for each sample in x.

    Args:
    - x (ndarray): an array of shape (batch_size, feat_dim) containing the feature vectors.
    - pdfs (list): a list of multivariate normal distributions, one for each class.
    - conf (SimpleNamespace): an instance of the SimpleNamespace class containing the following keys:
        - num_classes (int): the number of classes.
        - batch_size (int): the number of samples.
        - feat_dim (int): the feature dimension.

    Returns:
    - llrm (ndarray): an array of shape (batch_size, num_classes, num_classes) containing the log-likelihood ratio
      matrix for each sample in x.
    """

    likelihood = np.zeros((conf.batch_size, conf.num_classes))
    for cls_i in range(conf.num_classes):
        likelihood[:, cls_i] = np.log(pdfs[cls_i].pdf(x))

    llrm = np.zeros((conf.batch_size, conf.num_classes, conf.num_classes))
    for cls_i in range(conf.num_classes):
        for cls_j in range(conf.num_classes):
            # diagonal is zero by definition
            llrm[:, cls_i, cls_j] = likelihood[:, cls_i] - likelihood[:, cls_j]

    return llrm


def generate_likelihood_ratio_matrix(conf):
    """
    Generate sequential multivariate Gaussian likelihood ratio matrix for
    sequential density ratio estimation (SDRE).

    Args:
    - config (dict): contains following parameters,
        FEAT_DIM
        DENSITY_OFFSET
        BATCH_SIZE
        NUM_ITER
        TIME_STAMP
        NUM_CLASSES

    Returns:
    - x_iter_pool: numpy array of size (BATCH_SIZE * NUM_ITER, TIME_STAMP, FEAT_DIM)
                    SDRE data.
    - y_iter_pool: numpy array of size (BATCH_SIZE * NUM_ITER).
                    Represents ground-truth labels
    - llrm_iter_pool:  numpy array of size (BATCH_SIZE * NUM_ITER, TIME_STAMP, NUM_CLASSES, NUM_CLASSES).
                    Ground-truth LLR matrix.
    """

    meanvecs, covmat, pdfs = initialize_multivariate_gaussian(conf)

    x_iter_pool = []
    y_iter_pool = []
    llrm_iter_pool = []
    for iter_i in range(conf.num_iter):
        logger.info(f"Starting {iter_i=} / {conf.num_iter - 1}")
        x_cls_pool = []
        y_cls_pool = []
        llrm_cls_pool = []
        for cls_i in range(conf.num_classes):
            y = cls_i * np.ones((conf.batch_size))

            x_time_pool = []
            llrm_time_pool = []
            for t_i in range(conf.time_steps):
                x = np.random.multivariate_normal(
                    meanvecs[cls_i], covmat, conf.batch_size
                ).astype("float32")
                llrm = compute_log_likelihood_ratio_matrix(x, pdfs, conf)

                x_time_pool.append(x)
                llrm_time_pool.append(llrm)

            # reshape into (BATCH_SIZE, TIME_STEPS, FEAT_DIM)
            x_cls = np.stack(x_time_pool, axis=1)
            # reshape into (BATCH_SIZE, TIME_STEPS, NUM_CLASSES, NUM_CLASSES)
            llrm_cls = np.stack(llrm_time_pool, axis=1)
            assert x_cls.shape == (conf.batch_size, conf.time_steps, conf.feat_dim)
            assert y.shape == (conf.batch_size,)  # size y: (BATCH_SIZE)
            assert llrm_cls.shape == (
                conf.batch_size,
                conf.time_steps,
                conf.num_classes,
                conf.num_classes,
            )
            x_cls_pool.append(x_cls)
            y_cls_pool.append(y)
            llrm_cls_pool.append(llrm_cls)

        # reshape into (NUM_CLASSES * BATCH_SIZE, TIME_STEPS, FEAT_DIM)
        x_iter = np.concatenate(x_cls_pool, axis=0)
        # reshape into (NUM_CLASSES * BATCH_SIZE)
        y_iter = np.concatenate(y_cls_pool, axis=0)
        # reshape into (NUM_CLASSES * BATCH_SIZE, TIME_STEPS, NUM_CLASSES, NUM_CLASSES)
        llrm_iter = np.concatenate(llrm_cls_pool, axis=0)
        assert x_iter.shape == (
            conf.num_classes * conf.batch_size,
            conf.time_steps,
            conf.feat_dim,
        )
        assert y_iter.shape == (conf.num_classes * conf.batch_size,)
        assert llrm_iter.shape == (
            conf.num_classes * conf.batch_size,
            conf.time_steps,
            conf.num_classes,
            conf.num_classes,
        )
        x_iter_pool.append(x_iter)
        y_iter_pool.append(y_iter)
        llrm_iter_pool.append(llrm_iter)

    # reshape into (NUM_ITER * NUM_CLASSES * BATCH_SIZE, TIME_STEPS, FEAT_DIM)
    x_iter_pool = np.concatenate(x_iter_pool, axis=0)
    # reshape into (NUM_ITER * NUM_CLASSES * BATCH_SIZE)
    y_iter_pool = np.concatenate(y_iter_pool, axis=0)
    # reshape into (NUM_ITER * NUM_CLASSES * BATCH_SIZE, TIME_STEPS, NUM_CLASSES, NUM_CLASSES)
    llrm_iter_pool = np.concatenate(llrm_iter_pool, axis=0)
    assert x_iter_pool.shape == (
        conf.num_iter * conf.num_classes * conf.batch_size,
        conf.time_steps,
        conf.feat_dim,
    )
    assert y_iter_pool.shape == (conf.num_iter * conf.num_classes * conf.batch_size,)
    assert llrm_iter_pool.shape == (
        conf.num_iter * conf.num_classes * conf.batch_size,
        conf.time_steps,
        conf.num_classes,
        conf.num_classes,
    )

    logger.success("successfully generated SDRE data!")

    # accumulate evidence
    llrm_iter_pool = np.cumsum(llrm_iter_pool, axis=1)

    if conf.is_shuffle:
        logger.info("shuffling the data...")
        total_data = conf.num_iter * conf.num_classes * conf.batch_size
        dice = np.random.permutation(total_data)

        x_iter_pool = x_iter_pool[dice]
        y_iter_pool = y_iter_pool[dice]
        llrm_iter_pool = llrm_iter_pool[dice]

    logger.success("done and dusted!")
    return x_iter_pool, y_iter_pool, llrm_iter_pool


def sequential_slice(x, y, order_sprt):
    """Slice, copy, and concat a batch to make a time-sliced, augumented batch
    Effective batch size will be batch * (time_steps - order_sprt)).
    e.g., nosaic MNIST and 2nd-order SPRT:
        effective batch size is (20-2)=18 times larger than the original batch size.
    Args:
        x: A Tensor with shape (batch, time_steps, feature dimension).
        y: A Tensor with shape (batch).
        order_sprt: An int. The order of SPRT.
    Returns:
        x_slice: A Tensor with shape (batch*(time_steps-order_sprt), order_sprt+1, feat dim).
        y_slice: A Tensor with shape (batch*(time_steps-order_sprt),).
    Remark:
        y_slice may be a confusing name, because we copy and concatenate original y to obtain y_slice.
    """
    x, y = numpy_to_torch(x, y)

    time_steps = x.shape[1]

    if time_steps < order_sprt + 1:
        raise ValueError(
            "order_sprt must be <= time_steps - 1. Now order_sprt={}, time_steps={} .".format(
                order_sprt, time_steps
            )
        )

    for i in range(time_steps - order_sprt):
        if i == 0:
            x_slice = x[:, i : i + order_sprt + 1, :]
            y_slice = y
        else:
            x_slice = torch.cat([x_slice, x[:, i : i + order_sprt + 1, :]], 0)
            y_slice = torch.cat([y_slice, y], 0)

    return x_slice, y_slice


def sequential_slice_data(inputs, order_sprt):
    inputs = numpy_to_torch(inputs)

    time_steps = inputs.shape[1]

    if time_steps < order_sprt + 1:
        raise ValueError(
            "order_sprt must be <= time_steps - 1."
            f" Now {order_sprt=}, {time_steps=}."
        )

    for i in range(time_steps - order_sprt):
        if i == 0:
            x_slice = inputs[:, i : i + order_sprt + 1, :]
        else:
            x_slice = torch.cat([x_slice, inputs[:, i : i + order_sprt + 1, :]], 0)

    return x_slice


def sequential_slice_labels(labels, time_steps, order_sprt):
    labels = numpy_to_torch(labels)

    if time_steps < order_sprt + 1:
        raise ValueError(
            "order_sprt must be <= time_steps - 1."
            f" Now {order_sprt=}, {time_steps=}."
        )

    for i in range(time_steps - order_sprt):
        if i == 0:
            y_slice = labels
        else:
            y_slice = torch.cat([y_slice, labels], 0)

    return y_slice


def sequential_concat_logits(
    logits_slice: torch.Tensor, time_steps: int
) -> torch.Tensor:
    """
    Opposite operation of sequential_slice.
    logits_slice's shape will change
    from (batch * (time_steps - order_sprt), order_sprt + 1, feat dim )
    to  (batch, (time_steps - order_sprt), order_sprt + 1, feat dim).
    Args:
        logits_slice: A Tensor with shape (batch * (time_steps - order_sprt), order_sprt + 1, feat dim). This is the output of models.backbones_lstm.LSTMModel.__call__(inputs, training).
        time_steps: An int. 20 for nosaic MNIST.
        y_slice: A Tensor with shape (batch*(time_steps - order_sprt),). Default: None.
    Returns:
        A Tensor with shape (batch, (time_steps - order_sprt), order_sprt + 1, feat dim).
        If y_slice is not None, also returns a Tensor with shape (batch).
    """

    order_sprt = int(logits_slice.shape[1] - 1)
    batch_size = int(logits_slice.shape[0] / (time_steps - order_sprt))
    feat_dim = logits_slice.shape[-1]

    batch_size = logits_slice.shape[0] // (time_steps - logits_slice.shape[1] + 1)
    feat_dim = logits_slice.shape[-1]

    x_concat = logits_slice.reshape(
        time_steps - order_sprt, batch_size, order_sprt + 1, feat_dim
    )
    # (batch, duration - order_sprt, order_sprt + 1, feat_dim)
    x_concat = x_concat.permute(1, 0, 2, 3)

    return x_concat


def sequential_concat_labels(
    labels_slice: torch.Tensor, batch_size: int
) -> torch.Tensor:
    """ """
    y_concat = labels_slice[:batch_size]
    return y_concat


def reshape_for_featext(x, y, feat_dim):
    """(batch, time_steps) to (batch * time_steps,)"""
    x_shape = x.shape
    batch_size = x_shape[0]
    time_steps = x_shape[1]

    # disentangle
    x = x.reshape(-1, feat_dim[0], feat_dim[1], feat_dim[2])

    y = y.repeat(time_steps)
    y = y.reshape(time_steps, batch_size)
    y = y.transpose(0, 1)
    y = y.reshape(
        -1,
    )

    return x, y



from typing import Any, Dict, Tuple

import torch
import torch.optim as optim
from loguru import logger
from ..utils.misc import extract_params_from_config


def get_optimizer(conf):
    if "adam" in conf.optimizer.lower():
        return optim.AdamW
    elif conf.optimizer.lower() == "rmsprop":
        return optim.RMSprop
    else:
        raise ValueError(f'Optimizer "{conf.optimizer}" is not implemented!')


def initialize_optimizer(
    model: torch.nn.Module, config: Dict[str, Any]
) -> Tuple[torch.nn.Module, optim.Optimizer]:
    """
    Initializes the optimizer for training the model, based on the configuration parameters.

    Args:
    - model: The model to be trained.
    - config: A dictionary containing model configuration parameters.

    Returns:
    - Tuple[Module, optim.Optimizer]: A tuple containing the initialized model and optimizer.
    """
    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "LEARNING_RATE",
            "LR_DECAY_STEPS",
            "OPTIMIZER",
            "WEIGHT_DECAY",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    base_optimizer = get_optimizer(conf)

    optimizer = base_optimizer(
        model.parameters(),
        weight_decay=conf.weight_decay,
        lr=conf.learning_rate,
    )
    logger.info(f"Optimizer:\n{optimizer}")

    return model, optimizer



import os
import glob
from typing import List
from types import SimpleNamespace

import numpy as np
from termcolor import colored
from loguru import logger
import torch
from ..utils.misc import extract_params_from_config, convert_torch_to_numpy


@logger.catch
def initialize_objectives(config):
    """ """
    # check if necessary parameters are defined in the config file
    requirements = set(["OPTIMIZATION_CANDIDATES", "OPTIMIZATION_TARGET"])
    conf = extract_params_from_config(requirements, config)
    if "all" in conf.optimization_target.lower():
        num_candidates = len(conf.optimization_candidates)
        return tuple([np.Inf] * num_candidates)  # multiobjectives:
    else:
        return tuple([np.Inf])


def finalize_objectives(best):
    """
    Args:
    -best (tuple): objective(s) to be minimized.
    """

    if len(best) == 1:
        return best[0]
    elif len(best) > 1:
        best0, *remaining_bests = best
        return best0, *remaining_bests


def update_and_save_result(
    model, optimizer, config, best, performance_metrics, global_step
):
    """
    Update the best result according to the optimization target.

    Args:
    - config (dict): Configuration dictionary that contains the optimization target.
    - ckpt_manager: (TensorFlow checkpoint manager)
    - best (tuple): current best values of the objective
    - performance_metrics (dict): dictionary containing mean macro-average recall (macrec),
    - global_step (int): Current iteration number.
    Returns:
    - best (tuple): The updated best value.
    """

    def concatenate_variables(variables: List, n_digits: int) -> str:
        """ """

        def format_float(f, n_digits):
            return "{{:.{}f}}".format(n_digits).format(
                np.round(f * 10**n_digits) / 10**n_digits
            )

        # make double sure they're numpy
        variables = convert_torch_to_numpy(variables)
        variables = [format_float(v, n_digits) for v in variables]
        return "_".join(variables)

    def save_checkpoint(model, optimizer, best, conf, global_step):
        # save a checkpoint
        model_path = (
            f"{conf.dir_ckptlogs}/"
            f"ckpt_step{global_step}_target_"
            f"{conf.optimization_target}{concatenate_variables(best, n_digits=4)}"
            f".pt"
        )

        torch.save(
            {
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
            },
            model_path,
        )

        logger.info(colored("Saved checkpoint", "cyan") + f" at step {global_step}.")

    def keep_max_num_saved_models(conf: SimpleNamespace, max_to_keep: int = 1) -> None:
        """
        This function ensures that PyTorch saves a maximum number of checkpoint files (.pth format) in a given directory.
        If there are more saved models than the specified limit, it will delete the oldest ones, based on their creation time.

        Args:
        - conf (dict): stores conf.ckptlogs to specify checkpoint log directory.
        - max_to_keep (int, optional): The maximum number of checkpoint files to save in the directory. Defaults to 1.
        """
        # Get a list of all files in the directory
        models = glob.glob(f"{conf.dir_ckptlogs}/" + "*.pt")

        # Sort the files based on their creation time
        models = sorted(models, key=os.path.getmtime)
        if len(models) > max_to_keep:
            os.remove(models[0])
            logger.info(f"Removed the oldest model for {max_to_keep=}")

    def targets_sanity_check(targets, conf_targets) -> dict:
        """
        Check the optimization candidates defined in the config file and optimization targets match.
        Sort the order of targets if it is incongruent with that of the config file.

        Args:
        - targets (dict): optuna optimization targets. Note that all targets will be optimized
                          when config['OPTIMIATION_TARGET] is 'all.'
        - conf_targets (list): optimization targets listed in the config file.

        Returns:
        - sorted_targets (dict): sorted targets according to the order of conf_targets.
        """
        # Assert the number of key-value pairs of the two dictionaries are equal
        assert len(conf_targets) == len(targets)

        # Check if a set of values of 'conf' are same as a set of keys of 'target'
        assert set(conf_targets) == set(targets.keys())

        # Sort the order of target so that 'conf's keys and target's values are in the same order
        sorted_targets = {}
        for key in conf_targets:
            sorted_targets[key] = targets[key]
        return sorted_targets

    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "OPTIMIZATION_CANDIDATES",
            "OPTIMIZATION_TARGET",
            "DIR_CKPTLOGS",
            "MAX_TO_KEEP",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    # Optuna *minimize* the objective by default - give it an error, not accuracy or equivalent
    targets = {
        # MacRec should be larger... flip the sign
        "macrec": 1 - performance_metrics["mean_macro_recall"],
        # MABS should be smaller.
        "mabs": performance_metrics["mean_abs_error"],
        # AUSAT curve from confmx has accuracy as y-axis: flip it
        "ausat_confmx": 1 - performance_metrics["ausat_from_confmx"],
    }

    targets = targets_sanity_check(targets, conf.optimization_candidates)

    if "all" in conf.optimization_target.lower():
        target = tuple(targets.values())

    elif conf.optimization_target.lower() in targets.keys():
        target = tuple([targets.get(conf.optimization_target.lower())])

    else:
        raise ValueError("Unknown optimization target!")

    indices = [
        i for i, (x, y) in enumerate(zip(target, best)) if x < y
    ]  # smaller is better
    if global_step == 0:
        best = target

    elif indices:  # update the best value!
        best = tuple([target[i] if i in indices else x for i, x in enumerate(best)])
        logger.info(colored("Best value updated!", "cyan"))

        save_checkpoint(model, optimizer, best, conf, global_step)

        keep_max_num_saved_models(conf, conf.max_to_keep)

    return best



import math
import os
from typing import Tuple

import pyfiglet
import torch
from loguru import logger
from termcolor import colored
from torch import Tensor, nn
from torch.nn import TransformerEncoder, TransformerEncoderLayer
from torchinfo import summary

from ..models.losses import calc_scores, compute_llrs, restore_lost_significance
from ..utils.data_processing import sequential_concat_logits, sequential_slice_data
from ..utils.hyperparameter_tuning import format_model_name
from ..utils.misc import extract_params_from_config
from ..utils.performance_metrics import (
    check_diag_zeros_positive_offdiag,
    thresh_truncated_MSPRT,
    threshold_generator,
    tile_constant_threshold,
)


def import_model(config, device, tb_writer=None):
    """
    Select LSTM- or TRANSFORMER- based temporal integrator.
    If IS_RESUME=True, load a model from saved checkpoint.

    Args:
    - config (dict): dictionary that specifies model backbone.
    - device (torch.device): specifies computation device to which data and model are moved.
    - tb_writer (torch.utils.tensorboard.SummaryWriter): optional, provide to add a network graph to TensorBoard.

    Returns:
    - model (PyTorch model)
    """

    # check if necessary parameters are defined in the config file
    requirements = set(["MODEL_BACKBONE", "IS_COMPILE", "MODE"])
    conf = extract_params_from_config(requirements, config)

    model_name = format_model_name(conf.model_backbone)

    if "LSTM" in model_name:
        model = B2BsqrtTANDEM(config)
    elif "TRANSFORMER" in model_name:
        model = TANDEMformer(config)
    else:
        raise ValueError(f"Unsupported model {conf.model_backbone} found!")

    # print model layers and #parameters
    model.summarize(tb_writer, device)
    # Note that model.to(device) is actually redundant here because
    # torchinfo.summary moves the model to device. Keeping it for educational purpose.
    model.to(device)
    logger.info("model moved onto: " + colored(f"{device}.", "yellow"))

    # compile for optimization
    if conf.is_compile:
        logger.info("compiled the model for optimization.")
        return torch.compile(model, mode=conf.mode)
    else:
        return model


def load_pretrained_states(model, optimizer, config) -> None:
    # check if necessary parameters are defined in the config file
    requirements = set(["IS_RESUME", "PATH_RESUME"])
    conf = extract_params_from_config(requirements, config)

    # If resume
    if conf.is_resume:
        assert os.path.exists(
            conf.path_resume
        ), f"Checkpoint does not exist: {conf.path_resume}"

        checkpoint = torch.load(conf.path_resume)
        model.load_state_dict(checkpoint["model_state_dict"])
        optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

        logger.info(f"model and optimizer states were loaded from: {conf.path_resume}")
    else:
        return


# custom LSTM to accept new activation functions
class LSTM(nn.Module):
    """
    PyTorch implementation of a Long Short-Term Memory (LSTM) module.

    Args:
        feat_dim (int): Number of input features.
        hidden_size (int): Number of hidden units in the LSTM layer.

    Inputs:
        x (torch.Tensor): Input tensor of shape (batch_size, time_steps, feat_dim).

    Outputs:
        output (torch.Tensor): Output tensor of shape (batch_size, time_steps, hidden_size).

    References:
        - Hochreiter, S., & Schmidhuber, J. (1997). Long short-term memory. Neural computation, 9(8), 1735-1780.
    """

    def __init__(
        self,
        feat_dim,
        hidden_size,
        input_activation=nn.Sigmoid(),
        output_activation=nn.Tanh(),
    ):
        """
        Initializes the LSTM layer.

        Args:
            feat_dim (int): Number of input features.
            hidden_size (int): Number of hidden units in the LSTM layer.
            input_activation (callable): Input activation function a.k.a. recurrent activation function (default: torch.sigmoid).
            output_activation (callable): Output activation function (default: torch.tanh).
        """
        super(LSTM, self).__init__()
        self.feat_dim = feat_dim
        self.hidden_size = hidden_size
        self.input_activation = input_activation
        self.output_activation = output_activation

        # Initialize weight matrices and bias vectors
        self.W_i = nn.Parameter(torch.Tensor(feat_dim, hidden_size))
        self.U_i = nn.Parameter(torch.Tensor(hidden_size, hidden_size))
        self.b_i = nn.Parameter(torch.Tensor(hidden_size))

        self.W_f = nn.Parameter(torch.Tensor(feat_dim, hidden_size))
        self.U_f = nn.Parameter(torch.Tensor(hidden_size, hidden_size))
        self.b_f = nn.Parameter(torch.Tensor(hidden_size))

        self.W_c = nn.Parameter(torch.Tensor(feat_dim, hidden_size))
        self.U_c = nn.Parameter(torch.Tensor(hidden_size, hidden_size))
        self.b_c = nn.Parameter(torch.Tensor(hidden_size))

        self.W_o = nn.Parameter(torch.Tensor(feat_dim, hidden_size))
        self.U_o = nn.Parameter(torch.Tensor(hidden_size, hidden_size))
        self.b_o = nn.Parameter(torch.Tensor(hidden_size))

        self.init_weights()

    def init_weights(self):
        """
        Initializes the weights and biases of the LSTM layer using Xavier initialization.
        """
        # Initialize weights and biases using Xavier initialization
        nn.init.xavier_uniform_(self.W_i)
        nn.init.xavier_uniform_(self.U_i)
        nn.init.constant_(self.b_i, 0)

        nn.init.xavier_uniform_(self.W_f)
        nn.init.xavier_uniform_(self.U_f)
        nn.init.constant_(self.b_f, 0)

        nn.init.xavier_uniform_(self.W_c)
        nn.init.xavier_uniform_(self.U_c)
        nn.init.constant_(self.b_c, 0)

        nn.init.xavier_uniform_(self.W_o)
        nn.init.xavier_uniform_(self.U_o)
        nn.init.constant_(self.b_o, 0)

    def forward(self, x: Tensor) -> Tensor:
        """
        Passes the input tensor through the LSTM layer.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, time_steps, feat_dim).

        Returns:
            output (torch.Tensor): Output tensor of shape (batch_size, time_steps, hidden_size).
        """
        # Initialize hidden and cell state tensors
        batch_size, time_steps, _ = x.shape
        h_t = torch.zeros(batch_size, self.hidden_size).to(x.device)
        c_t = torch.zeros(batch_size, self.hidden_size).to(x.device)

        # Iterate over time steps
        output = []
        for t in range(time_steps):
            x_t = x[:, t, :]

            # Input gate
            i_t = self.input_activation(
                torch.matmul(x_t, self.W_i) + torch.matmul(h_t, self.U_i) + self.b_i
            )

            # Forget gate
            f_t = self.input_activation(
                torch.matmul(x_t, self.W_f) + torch.matmul(h_t, self.U_f) + self.b_f
            )

            # Output gate
            o_t = self.input_activation(
                torch.matmul(x_t, self.W_o) + torch.matmul(h_t, self.U_o) + self.b_o
            )

            # Cell state
            c_tilda_t = self.output_activation(
                torch.matmul(x_t, self.W_c) + torch.matmul(h_t, self.U_c) + self.b_c
            )
            c_t = f_t * c_t + i_t * c_tilda_t

            h_t = o_t * self.output_activation(c_t)

            output.append(h_t.unsqueeze(1))

        # Stack output tensors and return
        return torch.cat(output, dim=1)


class PositionalEncoding(nn.Module):
    """
    A PyTorch module that applies a learnable positional encoding to the input tensor.
    Adapted from the official document:
        https://pytorch.org/tutorials/beginner/transformer_tutorial.html

    Attributes:
        dropout: nn.Dropout, the dropout layer applied to the output tensor.
        feat_dim: int, the number of features in the input tensor.
        pe: torch.Tensor, the positional encoding tensor.

    Methods:
        forward: Applies the positional encoding to the input tensor and returns the output.

    """

    pe: Tensor  # type hinting for x = x + self.pe[: x.size(0)]

    def __init__(self, feat_dim: int, dropout: float = 0.1, max_len: int = 5000):
        """
        Initializes a new PositionalEncoding module.

        Args:
            feat_dim: int, the number of features in the input tensor.
            dropout: float, the dropout probability for the output tensor.
            max_len: int, the maximum length of the input sequence.
        """
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.feat_dim = feat_dim
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, feat_dim, 2) * (-math.log(10000.0) / feat_dim)
        )
        pe = torch.zeros(max_len, 1, feat_dim)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        # register the pe tensor as a buffer, which means that it will be included in the model's
        # state dictionary but won't be considered as a model parameter that needs to be optimized
        # during training.
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor, batch_first: bool = True) -> Tensor:
        """
        Applies the positional encoding to the input tensor and returns the output.

        Args:
            x: Tensor, the input tensor with shape [time_steps, batch_size, embedding_dim] if batch_first=False
                   else [batch_size, time_steps, embedding_dim] if batch_first=True
            batch_first: bool, whether the input tensor has batch_size as its first dimension.

        Returns:
            Tensor, the output tensor with the positional encoding added.
        """
        assert (
            x.shape[-1] == self.feat_dim
        ), f"Expected last dimension of x to be {self.feat_dim}, but got {x.shape[-1]} instead"
        assert (
            x.ndim == 3
        ), f"Expecting a Tensor with dimensions batch_size, time_steps, and embedding_dim but got {x.shape=}."
        x = x.permute(1, 0, 2) if batch_first else x
        x = x + self.pe[: x.size(0)]
        x = self.dropout(x)
        return x.permute(1, 0, 2) if batch_first else x


# custom activation functions
def b2bsqrt(x, alpha=1.0):
    alpha = torch.tensor(alpha)
    return torch.sign(x) * (torch.sqrt(alpha + torch.abs(x)) - torch.sqrt(alpha))


def b2bsqrtv2(x, alpha=1.0, beta=0.01):
    alpha = torch.tensor(alpha)
    beta = torch.tensor(beta)
    # add a linear component for anti-gradient vanishing
    return torch.sign(x) * (
        torch.sqrt(alpha + torch.abs(x)) + beta * torch.abs(x) - torch.sqrt(alpha)
    )


def b2bcbrt(x, alpha=1.0, gamma=2.0):
    alpha = torch.tensor(alpha)
    gamma = torch.tensor(gamma)
    return torch.sign(x) * (
        torch.pow(alpha + torch.abs(x), 2 / 3) - torch.pow(alpha, 2 / 3)
    )


def b2bexp(x, alpha=0.03, beta=0.1, tau=1000):
    alpha = torch.tensor(alpha)
    beta = torch.tensor(beta)
    tau = torch.tensor(tau)
    return torch.sign(x) * (
        -alpha * torch.exp(-torch.abs(x) / tau) + beta * torch.abs(x) + alpha
    )


def tanhplus(x, alpha=10.0, beta=0.02, tau=100.0):
    alpha = torch.tensor(alpha)
    beta = torch.tensor(beta)
    tau = torch.tensor(tau)
    # add a linear component for anti-gradient vanishing
    return alpha * torch.tanh(x / tau) + beta * x


def dullrelu(x, beta=0.05):
    beta = torch.tensor(beta)
    return beta * torch.maximum(torch.zeros_like(x), x)


def b2blog(x):
    return torch.sign(x) * (torch.log(1 + torch.abs(x)))


activations = {
    "b2bsqrt": b2bsqrt,
    "b2bsqrtv2": b2bsqrtv2,
    "b2bcbrt": b2bcbrt,
    "b2bexp": b2bexp,
    "tanhplus": tanhplus,
    "dullrelu": dullrelu,
    "b2blog": b2blog,
    "relu": nn.ReLU(),
    "gelu": nn.GELU(),
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "sine": torch.sin,
}


def get_activation(activation: str):
    """
    Returns the activation function specified by the argument `activation`.

    Parameters:
    - activation (str): The name of the activation function. Supported values are:
    'b2bsqrt', 'b2bcbrt', 'b2bexp', 'tanhplus', 'dullrelu', 'B2Blog', 'relu', 'sigmoid', and 'tanh'.

    Returns:
    - Callable: The specified activation function.

    Raises:
        ValueError: If the specified activation function is not supported.
    """
    if activation.lower() not in activations:
        raise ValueError(f"{activation} activation function is not supported!")
    return activations[activation.lower()]


class BaseTANDEM(nn.Module):
    """
    Base PyTorch model that defines utility functions that are common to all models.

    This class inherits from `nn.Module`. All TANDEM models should inherit from this class
    instead of directly inheriting from `nn.Module` to make use of the `summarize` function.
    """

    def __init__(self, config, feature_size):
        super().__init__()

        # check if necessary parameters are defined in the config file
        requirements = set(
            [
                "NUM_CLASSES",
                "TIME_STEPS",
                "FEAT_DIM",
                "ORDER_SPRT",
                "ACTIVATION_FC",
                "BATCH_SIZE",
                "OBLIVIOUS",
                "NUM_THRESH",
                "SPARSITY",
                "DROPOUT",
                "IS_NORMALIZE",
                "IS_ADAPTIVE_LOSS",
                "IS_POSITIONAL_ENCODING",
                "IS_TRAINABLE_ENCODING",
                "OPTIMIZATION_CANDIDATES",
            ]
        )
        conf = extract_params_from_config(requirements, config)

        self.num_classes = conf.num_classes
        self.time_steps = conf.time_steps
        self.feat_dim = conf.feat_dim
        self.order_sprt = conf.order_sprt
        self.batch_size = conf.batch_size
        self.oblivious = conf.oblivious
        self.dropout = nn.Dropout(p=conf.dropout)
        self.num_thresh = conf.num_thresh
        self.sparsity = conf.sparsity

        # number of elements in the upper triangle (kC2 where k is the num_classes)
        self.vec_dim = self.num_classes * (self.num_classes - 1) // 2

        self.is_normalize = conf.is_normalize
        if self.is_normalize:
            self.layer_norm = nn.LayerNorm(feature_size)

        self.is_positional_encoding = conf.is_positional_encoding
        self.is_trainable_encoding = conf.is_trainable_encoding
        if self.is_trainable_encoding:
            self.trainable_pos_encoding = nn.Parameter(
                torch.randn(self.order_sprt + 1, self.feat_dim)
            )
        else:
            self.pos_encoder = PositionalEncoding(conf.feat_dim, conf.dropout)

        # for debug. delete this later
        self.pos_encoder = PositionalEncoding(conf.feat_dim, conf.dropout)

        if conf.is_adaptive_loss:
            self.adaptive_loss_weights = nn.Parameter(
                torch.randn(len(conf.optimization_candidates))
            )

        # FC layers
        self.activation_logit = get_activation(conf.activation_fc)
        self.fc_logits = nn.Linear(
            in_features=feature_size, out_features=conf.num_classes, bias=True
        )

    def normalized_sum_pool(self, x: Tensor, max_time_steps: int) -> Tensor:
        """
        NSP layer for precise sequential density ratio estimation.
        (Ebihara+, ICASSP 2023)

        Args:
        -x (Tensor): size (effective_batch_size, effective_time_steps, feat_dim)
         effective_time_steps \in {1, max_time_steps}
        -max_time_steps (int): maximum time_steps to consider. e.g) order_sprt + 1.

        Return:
        torch.Tensor of the size size as x.
        """
        return torch.sum(x, dim=1) / max_time_steps

    def features_to_logits(self, x, time_steps):
        """
        Args:
            x: A Tensor. Output of network with shape=(batch, time_steps, self.width_lstm).
        Returns:
        - logit_stack (PyTorch Tensor): (batch, time_steps, num_classes)
        """
        logit_stack = []
        for i in range(time_steps):
            y = x[:, i, :]
            if self.is_normalize:
                y = self.layer_norm(y)
            # final FC layer
            y_class = self.activation_logit(y)
            y_class = self.fc_logits(y)
            logit_stack.append(y_class)

        logit_stack = torch.stack(logit_stack, dim=1)
        return logit_stack

    def logits_to_llrs(self, logits: torch.Tensor, time_steps: int) -> torch.Tensor:
        """
        Args:
            logits (PyTorch Tensor): (effective_batch_size, effective_time_steps, num_classes)
            time_steps (int): full-length time_steps >= effective_time_steps.


        """
        logits_concat = sequential_concat_logits(logits, time_steps=time_steps)
        llrs = compute_llrs(logits_concat, oblivious=self.oblivious)

        return llrs

    def markov_time_slicer(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, int, int]:
        """
        Args:
        - inputs (PyTorch Tensor): (batch_size, time_steps, feat_dim)

        Returns:
        - x_slice (PyTorch Tensor): (batch_size * (time_steps - order_sprt), order_sprt + 1, feat_dim)
        - eff_batch_size (int): effective batch size
        - eff_time_steps (int): effective time steps
        """

        x_slice = sequential_slice_data(inputs, self.order_sprt)

        shapes = x_slice.shape
        eff_batch_size = shapes[0]
        eff_time_steps = shapes[1]

        return x_slice, eff_batch_size, eff_time_steps

    def create_threshold(self, llrs: torch.Tensor) -> torch.Tensor:
        """ """
        thresh = threshold_generator(
            llrs, self.num_thresh, self.sparsity
        )  # (num_thresh,)

        thresh = tile_constant_threshold(
            thresh, self.batch_size, self.time_steps, self.num_classes
        )

        thresh = thresh_truncated_MSPRT(thresh)

        check_diag_zeros_positive_offdiag(thresh)

        return thresh

    def distance_between_llrs_and_thresholds(
        self, llrs: torch.Tensor, thresh: torch.Tensor
    ) -> torch.Tensor:
        """ """
        # Change LLRs and threshold scales
        # llrs, thresh = normalize_LLR_thresh_scale(llrs, thresh)
        # Otherwise margins_up and margins_low may contain lots of zeros.
        # No shape changes.
        # llrs: (batch, time_steps, num classes, num classes)
        # thresh: (num_thresh, time_steps, num classes, num classes)
        # Comment 20221201: May cause optimization instability because of the fluctuating LLR scales?
        #                   Maybe stop_grad (somewhere) needed?

        # Restore lost significance with epsilon
        llrs = restore_lost_significance(llrs)
        # (batch, time_steps, num cls, num cls)
        # To avoid double hit due to the values exactly equal to 0
        # in scores or when doing truncation (LLRs of the last frame).

        # scores_full: llr_mtx - thresh_mtx,
        # size:(num_thresh, batch_size, time_steps, num_classes, num_classes)
        scores_full = calc_scores(llrs, thresh)

        return scores_full

    def features_to_final_scores(
        self, outputs, effective_time_steps
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """ """
        # Make logits
        logits = self.features_to_logits(outputs, time_steps=effective_time_steps)

        # compute log-likelihood ratio. Also convert to full-time steps
        llrs = self.logits_to_llrs(logits, self.time_steps)

        thresh = self.create_threshold(llrs)

        scores_full = self.distance_between_llrs_and_thresholds(llrs, thresh)

        return llrs, logits, thresh, scores_full

    def summarize(self, tb_writer, device):
        """ """
        logger.info(
            "\n" + pyfiglet.figlet_format("SPRT-    \nTANDEM", font="banner3-D")
        )
        logger.info(colored("official PyTorch implementation.", "blue"))

        self.eval()

        example_input = torch.randn(self.batch_size, self.time_steps, self.feat_dim).to(
            device
        )

        with torch.no_grad():
            # summary table by torchinfo
            # caution: this moves the model to the device
            sum_str = str(summary(self, example_input.shape, device=device, verbose=0))

            example_output = self.forward(example_input)

            # export to tensorboard
            if tb_writer:
                tb_writer.add_graph(self, example_input)
        example_output = (
            example_output[0] if isinstance(example_output, tuple) else example_output
        )

        logger.info("Network summary:\n" + sum_str)
        logger.info(f"Example input shape: {example_input.shape}")
        logger.info(f"Example output shape: {example_output.shape}")


class B2BsqrtTANDEM(BaseTANDEM):
    def __init__(self, config):
        # check if necessary parameters are defined in the config file
        requirements = set(
            [
                "WIDTH_LSTM",
                "ACTIVATION_OUTPUT",
                "ACTIVATION_INPUT",
            ]
        )
        conf = extract_params_from_config(requirements, config)

        super().__init__(config, feature_size=conf.width_lstm)

        # Parameters
        self.width_lstm = conf.width_lstm

        # activation functions
        self.activation_output = get_activation(conf.activation_output)
        self.activation_input = get_activation(conf.activation_input)

        self.rnn = LSTM(
            feat_dim=self.feat_dim,
            hidden_size=self.width_lstm,
            input_activation=self.activation_input,
            output_activation=self.activation_output,
        )

    def forward(self, inputs):
        """Calc logits.
        Args:
            inputs: A Tensor with shape=(batch, time_steps, feature dimension). E.g. (128, 20, 784) for nosaic MNIST.
            training: A boolean. Training flag used in BatchNormalization and dropout.
        Returns:
            outputs: A Tensor with shape=(batch, time_steps, num_classes).
        """
        # get sequence fragments 'x_slice', effective batch size, and effective time steps
        x_slice, eff_batch_size, eff_time_steps = self.markov_time_slicer(inputs)

        # Positinal encoding
        if self.is_positional_encoding:
            if self.is_trainable_encoding:
                final_inputs = x_slice + self.trainable_pos_encoding.unsqueeze(
                    0
                ).repeat(eff_batch_size, 1, 1)
            else:
                # (batch_size, time_steps, feat_dim)
                final_inputs = self.pos_encoder(x_slice, batch_first=True)
        else:
            final_inputs = x_slice
        # Feedforward
        # (batch_size, (effective) time_steps = order_sprt + 1, width_lstm)
        outputs = self.rnn(final_inputs)

        return self.features_to_final_scores(outputs, eff_time_steps)


class TANDEMformer(BaseTANDEM):
    """
    Transformer with the Normalized Summation Pooling (NSP) layer for precise SDRE.
    """

    def __init__(self, config):
        # check if necessary parameters are defined in the config file
        requirements = set(
            ["NUM_BLOCKS", "NUM_HEADS", "FF_DIM", "MLP_UNITS", "DROPOUT", "FEAT_DIM"]
        )
        conf = extract_params_from_config(requirements, config)

        super().__init__(config, feature_size=conf.mlp_units)

        self.decoder = nn.Linear(conf.feat_dim, conf.mlp_units)

        # main Transformer
        encoder_layers = TransformerEncoderLayer(
            conf.feat_dim,
            conf.num_heads,
            dim_feedforward=conf.ff_dim,
            dropout=conf.dropout,
            activation=self.activation_logit,
            batch_first=True,
            norm_first=False,
        )

        self.transformer_encoder = TransformerEncoder(
            encoder_layers, num_layers=conf.num_blocks
        )

        self.init_weights()

    def init_weights(self) -> None:
        initrange = 0.1
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)
        self.fc_logits.bias.data.zero_()
        self.fc_logits.weight.data.uniform_(-initrange, initrange)

    def transformer_classifier(self, inputs, max_time_steps):
        # pooling in time dimension: NSP layer
        inputs = self.normalized_sum_pool(inputs, max_time_steps)
        inputs = self.decoder(inputs)
        inputs = self.dropout(inputs)
        return inputs

    def meta_classifier(self, inputs, max_time_steps):
        # pooling in time dimension: NSP layer
        inputs = self.normalized_sum_pool(inputs, max_time_steps)
        inputs = self.dropout(inputs)
        return inputs

    def forward(
        self, inputs: Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Args:
            src: Tensor, shape [seq_len, batch_size]
            src_mask: Tensor, shape [seq_len, seq_len]

        Returns:
            output Tensor of shape [seq_len, batch_size, ntoken]
        """

        # get sequence fragments 'x_slice', effective batch size, and effective time steps
        x_slice, eff_batch_size, eff_time_steps = self.markov_time_slicer(inputs)

        outputs_pool = []
        for i in range(eff_time_steps):
            targinputs = x_slice[:, : i + 1, :]

            # Positinal encoding
            if self.is_positional_encoding:
                if self.is_trainable_encoding:
                    final_inputs = targinputs + self.trainable_pos_encoding.unsqueeze(
                        0
                    )[:, : i + 1, :].repeat(eff_batch_size, 1, 1)

                else:
                    final_inputs = self.pos_encoder(targinputs, batch_first=True)
            else:
                final_inputs = targinputs
            # mix the input matrix
            targoutputs = self.transformer_encoder(final_inputs)

            outputs_pool.append(
                self.transformer_classifier(targoutputs, max_time_steps=eff_time_steps)
            )
        # (batch_size, time_steps, mlp_feats)
        outputs = torch.stack(outputs_pool, dim=1)

        return self.features_to_final_scores(outputs, eff_time_steps)
